import pytest
import torch

from lgatr.layers.linear import EquiLinear
from lgatr.primitives.config import gatr_config
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", [(100,)])
@pytest.mark.parametrize("in_mv_channels, out_mv_channels", [(200, 5), (16, 16), (5, 200)])
@pytest.mark.parametrize(
    "in_s_channels, out_s_channels", [(None, None), (None, 100), (100, None), (32, 32)]
)
@pytest.mark.parametrize(
    "initialization", ["default", "small", "unit_scalar", "almost_unit_scalar"]
)
@pytest.mark.parametrize("use_fully_connected_subgroup", [True, False])
def test_linear_layer_initialization(
    initialization,
    batch_dims,
    in_mv_channels,
    out_mv_channels,
    in_s_channels,
    out_s_channels,
    use_fully_connected_subgroup,
    var_tolerance=10.0,
):
    """Tests the initialization of `EquiLinear`.

    The goal is that independent of the channel size, inputs with variance 1 are mapped to outputs
    with, very roughly, variance 1.
    """
    gatr_config.use_fully_connected_subgroup = use_fully_connected_subgroup

    # Create layer
    try:
        layer = EquiLinear(
            in_mv_channels,
            out_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=out_s_channels,
            initialization=initialization,
        )
    # Some initialization schemes are not implemented when data is all-scalar. That's fine.
    except NotImplementedError as exc:
        print(exc)
        return

    # Inputs
    inputs_mv = torch.randn(*batch_dims, in_mv_channels, 16)
    inputs_s = torch.randn(*batch_dims, in_s_channels) if in_s_channels is not None else None

    # Compute outputs
    outputs_mv, outputs_s = layer(inputs_mv, scalars=inputs_s)

    # Compute mean and variance of MV outputs
    mv_mean = outputs_mv[...].cpu().detach().to(torch.float64).mean(dim=(0, 1))
    mv_var = outputs_mv[...].cpu().detach().to(torch.float64).var(dim=(0, 1))

    print("Output multivector means and std by components:")
    for i, (mean_, var_) in enumerate(zip(mv_mean, mv_var, strict=False)):
        print(f"  Component {i}: mean = {mean_:.2f}, std = {var_**0.5:.2f}")

    # Check that the mean and variance agree with expectations
    if initialization == "default":
        target_mean = torch.zeros_like(mv_mean)
        target_var = torch.ones_like(mv_var) / 3.0  # Factor 3 comes from heuristics
    elif initialization == "small":
        target_mean = torch.zeros_like(mv_mean)
        target_var = 0.01 * torch.ones_like(mv_var) / 3.0
    elif initialization == "unit_scalar":
        target_mean = torch.zeros_like(mv_mean)
        target_mean[0] = 1.0
        if gatr_config.use_fully_connected_subgroup:
            target_mean[-1] = 1.0
        target_var = 0.01 * torch.ones_like(mv_var) / 3.0
    elif initialization == "almost_unit_scalar":
        target_mean = torch.zeros_like(mv_mean)
        target_mean[0] = 1.0
        if gatr_config.use_fully_connected_subgroup:
            target_mean[-1] = 1.0
        target_var = 0.25 * torch.ones_like(mv_var) / 3.0
    else:
        raise ValueError(initialization)

    assert torch.all(mv_mean > target_mean - 0.3)
    assert torch.all(mv_mean < target_mean + 0.3)
    assert torch.all(mv_var > target_var / var_tolerance)
    assert torch.all(mv_var < target_var * var_tolerance)

    # Same for scalar outputs
    if out_s_channels is not None:
        s_mean = outputs_s[...].cpu().detach().to(torch.float64).mean().item()
        s_var = outputs_s[...].cpu().detach().to(torch.float64).var().item()

        print(f"Output scalar: mean = {s_mean:.2f}, std = {s_var**0.5:.2f}")
        assert -0.3 < s_mean < 0.3
        if initialization in {"default", "unit_scalar", "almost_unit_scalar"}:
            assert 1.0 / 3.0 / var_tolerance < s_var < 1.0 / 3.0 * var_tolerance
        else:
            assert 0.01 / 3.0 / var_tolerance < s_var < 0.01 / 3.0 * var_tolerance

    # restore defaults
    gatr_config.use_fully_connected_subgroup = True


@pytest.mark.parametrize("rescaling", [0.0, -2.0, 100.0])
@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("in_mv_channels", [9, 1])
@pytest.mark.parametrize("out_mv_channels", [7, 1])
@pytest.mark.parametrize("in_s_channels", [None, 3])
@pytest.mark.parametrize("out_s_channels", [None, 4])
def test_linear_layer_linearity(
    batch_dims,
    in_mv_channels,
    out_mv_channels,
    in_s_channels,
    out_s_channels,
    rescaling,
):
    """Tests that the EquiLinear layer indeed describes a linear map (when the bias is deactivated).

    Checks that `f(x + rescaling * y) = f(x) + rescaling * f(y)` for random inputs `x`, `y` and
    linear layer `f(x)`.
    """
    layer = EquiLinear(
        in_mv_channels,
        out_mv_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        bias=False,
    )

    # Inputs
    x_mv = torch.randn(*batch_dims, in_mv_channels, 16)
    y_mv = torch.randn(*batch_dims, in_mv_channels, 16)
    xy_mv = x_mv + rescaling * y_mv

    if in_s_channels:
        x_s = torch.randn(*batch_dims, in_s_channels)
        y_s = torch.randn(*batch_dims, in_s_channels)
        xy_s = x_s + rescaling * y_s
    else:
        x_s, y_s, xy_s = None, None, None

    # Compute outputs
    o_xy_mv, o_xy_s = layer(xy_mv, scalars=xy_s)
    o_x_mv, o_x_s = layer(x_mv, scalars=x_s)
    o_y_mv, o_y_s = layer(y_mv, scalars=y_s)

    # Check equality
    torch.testing.assert_close(o_xy_mv, o_x_mv + rescaling * o_y_mv, **TOLERANCES)

    if out_s_channels is not None:
        torch.testing.assert_close(o_xy_s, o_x_s + rescaling * o_y_s, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("in_mv_channels", [9, 1])
@pytest.mark.parametrize("out_mv_channels", [7, 1])
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("in_s_channels", [None, 3])
@pytest.mark.parametrize("out_s_channels", [None, 4])
@pytest.mark.parametrize("use_fully_connected_subgroup", [True, False])
def test_linear_layer_equivariance(
    batch_dims,
    in_mv_channels,
    out_mv_channels,
    in_s_channels,
    out_s_channels,
    bias,
    use_fully_connected_subgroup,
):
    """Tests the equi_linear() primitive for equivariance."""
    gatr_config.use_fully_connected_subgroup = use_fully_connected_subgroup

    layer = EquiLinear(
        in_mv_channels,
        out_mv_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        bias=bias,
    )
    data_dims = tuple(list(batch_dims) + [in_mv_channels])
    scalars = None if in_s_channels is None else torch.randn(*batch_dims, in_s_channels)
    check_pin_equivariance(
        layer, 1, fn_kwargs=dict(scalars=scalars), batch_dims=data_dims, **TOLERANCES
    )

    # restore defaults
    gatr_config.use_fully_connected_subgroup = True



"""Embedding and extracting pseudoscalars into multivectors."""

import torch


def embed_pseudoscalar(pseudoscalars: torch.Tensor) -> torch.Tensor:
    """Embeds a pseudoscalar tensor into multivectors.

    Parameters
    ----------
    pseudoscalars: torch.Tensor
        Scalar inputs with shape (..., 1).

    Returns
    -------
    multivectors: torch.Tensor
        Multivector outputs with shape (..., 16).
        ``multivectors[..., [15]]`` is the same as ``pseudoscalars``. The other components are zero.
    """
    assert pseudoscalars.shape[-1] == 1
    non_scalar_shape = list(pseudoscalars.shape[:-1]) + [15]
    non_scalar_components = torch.zeros(
        non_scalar_shape, device=pseudoscalars.device, dtype=pseudoscalars.dtype
    )
    embedding = torch.cat((non_scalar_components, pseudoscalars), dim=-1)

    return embedding


def extract_pseudoscalar(multivectors: torch.Tensor) -> torch.Tensor:
    """Extracts pseudoscalar components from multivectors.

    Parameters
    ----------
    multivectors: torch.Tensor
        Multivector inputs with shape (..., 16).

    Returns
    -------
    pseudoscalars: torch.Tensor
        Pseudoscalar component of multivectors with shape (..., 1).
    """

    return multivectors[..., [15]]



import torch
from einops import rearrange
from torch import nn

from ..layer_norm import EquiLayerNorm
from ..linear import EquiLinear
from .config import SelfAttentionConfig


class QKVModule(nn.Module):
    """Compute (multivector and scalar) queries, keys, and values via multi-head attention.
    This class is only used in self-attention. We do it manually for cross-attention.

    Parameters
    ----------
    config: SelfAttentionConfig
        Attention configuration
    """

    def __init__(self, config: SelfAttentionConfig):
        super().__init__()
        self.in_linear = EquiLinear(
            in_mv_channels=config.in_mv_channels + config.additional_qk_mv_channels,
            out_mv_channels=3 * config.hidden_mv_channels * config.num_heads,
            in_s_channels=config.in_s_channels + config.additional_qk_s_channels,
            out_s_channels=(
                None
                if config.in_s_channels is None
                else 3 * config.hidden_s_channels * config.num_heads
            ),
        )
        self.norm_qkv = EquiLayerNorm()
        self.config = config

    def forward(
        self,
        inputs,
        scalars,
        additional_qk_features_mv=None,
        additional_qk_features_s=None,
    ):
        """Evaluate head-wise queries, keys, and values. The heads have size
        `head_mv_channels=mv_channels*increase_hidden_channels // num_heads` and
        `head_s_channels=s_channels*increase_hidden_channels // num_heads`.

        Parameters
        ----------
        inputs : torch.Tensor
            Multivector inputs with shape (..., items, mv_channels, 16)
        scalars : torch.Tensor
            Scalar inputs with shape (..., items, s_channels)
        additional_qk_features_mv : None or torch.Tensor
            Additional multivector features that for the Q/K computation
        additional_qk_features_s : None or torch.Tensor
            Additional scalar features for the Q/K computation

        Returns
        -------
        q_mv : torch.Tensor
            Multivector queries with shape (..., heads, items, head_mv_channels, 16)
        k_mv : torch.Tensor
            Multivector keys with shape (..., heads, items, head_mv_channels, 16)
        v_mv : torch.Tensor
            Multivector values with shape (..., heads, items, head_mv_channels, 16)
        q_s : torch.Tensor
            Scalar queries with shape (..., heads, items, head_s_channels)
        k_s : torch.Tensor
            Scalar keys with shape (..., heads, items, head_s_channels)
        v_s : torch.Tensor
            Scalar values with shape (..., heads, items, head_s_channels)
        """

        # Additional inputs
        if additional_qk_features_mv is not None:
            inputs = torch.cat((inputs, additional_qk_features_mv), dim=-2)
        if additional_qk_features_s is not None:
            scalars = torch.cat((scalars, additional_qk_features_s), dim=-1)

        qkv_mv, qkv_s = self.in_linear(
            inputs, scalars
        )  # (..., num_items, 3 * hidden_channels * num_heads, 16)
        qkv_mv = rearrange(
            qkv_mv,
            "... items (qkv hidden num_heads) x -> qkv ... num_heads items hidden x",
            num_heads=self.config.num_heads,
            hidden=self.config.hidden_mv_channels,
            qkv=3,
        )
        q_mv, k_mv, v_mv = qkv_mv  # each: (..., num_heads, num_items, num_channels, 16)

        # Same, for optional scalar components
        if qkv_s is not None:
            qkv_s = rearrange(
                qkv_s,
                "... items (qkv hidden num_heads) -> qkv ... num_heads items hidden",
                num_heads=self.config.num_heads,
                hidden=self.config.hidden_s_channels,
                qkv=3,
            )
            q_s, k_s, v_s = qkv_s  # each: (..., num_heads, num_items, num_channels)
        else:
            q_s, k_s, v_s = None, None, None

        q_mv, q_s = self.norm_qkv(q_mv, scalars=q_s)
        k_mv, k_s = self.norm_qkv(k_mv, scalars=k_s)
        v_mv, v_s = self.norm_qkv(v_mv, scalars=v_s)

        return q_mv, k_mv, v_mv, q_s, k_s, v_s


class MultiQueryQKVModule(nn.Module):
    """Compute (multivector and scalar) queries, keys, and values via multi-query attention.
    Compared to the QKVModule defined above, MultiQueryQKVModule shares keys and values
    across attention heads, which saves parameters.
    This class is only used in self-attention. We do it manually for cross-attention.

    Parameters
    ----------
    config: SelfAttentionConfig
        Attention configuration
    """

    def __init__(self, config: SelfAttentionConfig):
        super().__init__()

        # Q projection
        self.q_linear = EquiLinear(
            in_mv_channels=config.in_mv_channels + config.additional_qk_mv_channels,
            out_mv_channels=config.hidden_mv_channels * config.num_heads,
            in_s_channels=config.in_s_channels + config.additional_qk_s_channels,
            out_s_channels=config.hidden_s_channels * config.num_heads,
        )

        # Key and value projections (shared between heads)
        self.k_linear = EquiLinear(
            in_mv_channels=config.in_mv_channels + config.additional_qk_mv_channels,
            out_mv_channels=config.hidden_mv_channels,
            in_s_channels=config.in_s_channels + config.additional_qk_s_channels,
            out_s_channels=config.hidden_s_channels,
        )
        self.v_linear = EquiLinear(
            in_mv_channels=config.in_mv_channels,
            out_mv_channels=config.hidden_mv_channels,
            in_s_channels=config.in_s_channels,
            out_s_channels=config.hidden_s_channels,
        )
        self.norm_qkv = EquiLayerNorm()
        self.config = config

    def forward(
        self,
        inputs,
        scalars,
        additional_qk_features_mv=None,
        additional_qk_features_s=None,
    ):
        """Evaluate head-wise queries, keys, and values. The heads have size
        `head_mv_channels=mv_channels*increase_hidden_channels // num_heads` and
        `head_s_channels=s_channels*increase_hidden_channels // num_heads`.
        The keys and values are shared across heads, i.e. we set `num_heads=1`
        and later broadcast over heads.

        Parameters
        ----------
        inputs : torch.Tensor
            Multivector inputs with shape (..., items, mv_channels, 16)
        scalars : torch.Tensor
            Scalar inputs with shape (..., items, s_channels)
        additional_qk_features_mv : None or torch.Tensor
            Additional multivector features that should be provided for the Q/K computation (e.g.
            positions of objects)
        additional_qk_features_s : None or torch.Tensor
            Additional scalar features that should be provided for the Q/K computation (e.g.
            object types)

        Returns
        -------
        q_mv : torch.Tensor
            Multivector queries with shape (..., heads, items, head_mv_channels, 16)
        k_mv : torch.Tensor
            Multivector keys with shape (..., 1, items, head_mv_channels, 16)
        v_mv : torch.Tensor
            Multivector values with shape (..., 1, items, head_mv_channels, 16)
        q_s : torch.Tensor
            Scalar queries with shape (..., heads, items, head_s_channels)
        k_s : torch.Tensor
            Scalar keys with shape (..., 1, items, head_s_channels)
        v_s : torch.Tensor
            Scalar values with shape (..., 1, items, head_s_channels)
        """

        # Additional inputs
        if additional_qk_features_mv is not None:
            qk_inputs = torch.cat((inputs, additional_qk_features_mv), dim=-2)
        else:
            qk_inputs = inputs
        if scalars is not None and additional_qk_features_s is not None:
            qk_scalars = torch.cat((scalars, additional_qk_features_s), dim=-1)
        else:
            qk_scalars = scalars

        # Project to queries, keys, and values (multivector reps)
        q_mv, q_s = self.q_linear(
            qk_inputs, qk_scalars
        )  # (..., num_items, hidden_channels * num_heads, 16)
        k_mv, k_s = self.k_linear(qk_inputs, qk_scalars)  # (..., num_items, hidden_channels, 16)
        v_mv, v_s = self.v_linear(inputs, scalars)  # (..., num_items, hidden_channels, 16)

        # Rearrange to (..., heads, items, channels, 16) shape
        q_mv = rearrange(
            q_mv,
            "... items (hidden_channels num_heads) x -> ... num_heads items hidden_channels x",
            num_heads=self.config.num_heads,
            hidden_channels=self.config.hidden_mv_channels,
        )
        k_mv = rearrange(k_mv, "... items hidden_channels x -> ... 1 items hidden_channels x")
        v_mv = rearrange(v_mv, "... items hidden_channels x -> ... 1 items hidden_channels x")

        # Same for scalars
        if q_s is not None:
            q_s = rearrange(
                q_s,
                "... items (hidden_channels num_heads) -> ... num_heads items hidden_channels",
                num_heads=self.config.num_heads,
                hidden_channels=self.config.hidden_s_channels,
            )
            k_s = rearrange(k_s, "... items hidden_channels -> ... 1 items hidden_channels")
            v_s = rearrange(v_s, "... items hidden_channels -> ... 1 items hidden_channels")
        else:
            q_s, k_s, v_s = None, None, None

        q_mv, q_s = self.norm_qkv(q_mv, scalars=q_s)
        k_mv, k_s = self.norm_qkv(k_mv, scalars=k_s)
        v_mv, v_s = self.norm_qkv(v_mv, scalars=v_s)

        return q_mv, k_mv, v_mv, q_s, k_s, v_s



import pytest
import torch

from lgatr.layers import (
    ConditionalLGATrBlock,
    CrossAttentionConfig,
    MLPConfig,
    SelfAttentionConfig,
)
from tests.helpers import BATCH_DIMS, MILD_TOLERANCES, check_pin_equivariance

S_CHANNELS = [(3, 5), (2, 2)]


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("num_items,num_items_condition", [(2, 2), (2, 9)])
@pytest.mark.parametrize("mv_channels,mv_channels_condition", [(6, 6), (7, 11)])
@pytest.mark.parametrize("num_heads", [1, 4])
@pytest.mark.parametrize("s_channels,s_channels_condition", S_CHANNELS)
@pytest.mark.parametrize("dropout_prob", [None])
@pytest.mark.parametrize("multi_query_attention", [False, True])
def test_conditional_gatr_block_shape(
    batch_dims,
    num_items,
    num_items_condition,
    mv_channels,
    mv_channels_condition,
    num_heads,
    s_channels,
    s_channels_condition,
    multi_query_attention,
    dropout_prob,
):
    """Tests the output shape of ConditionalLGATrBlock."""
    inputs = torch.randn(*batch_dims, num_items, mv_channels, 16)
    scalars = None if s_channels is None else torch.randn(*batch_dims, num_items, s_channels)
    condition_mv = torch.randn(*batch_dims, num_items_condition, mv_channels_condition, 16)
    condition_s = (
        None
        if s_channels is None
        else torch.randn(*batch_dims, num_items_condition, s_channels_condition)
    )

    try:
        net = ConditionalLGATrBlock(
            mv_channels,
            s_channels=s_channels,
            condition_mv_channels=mv_channels_condition,
            condition_s_channels=s_channels_condition,
            attention=SelfAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            crossattention=CrossAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            mlp=MLPConfig(),
            dropout_prob=dropout_prob,
        )
    except NotImplementedError:
        # Some features require scalar inputs, and failing without them is fine
        return

    outputs, output_scalars = net(
        inputs,
        scalars=scalars,
        multivectors_condition=condition_mv,
        scalars_condition=condition_s,
    )

    assert outputs.shape == (*batch_dims, num_items, mv_channels, 16)
    if s_channels is not None:
        assert output_scalars.shape == (*batch_dims, num_items, s_channels)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("num_items,num_items_condition", [(2, 2), (2, 9)])
@pytest.mark.parametrize("mv_channels,mv_channels_condition", [(6, 6), (7, 11)])
@pytest.mark.parametrize("num_heads", [1, 4])
@pytest.mark.parametrize("s_channels,s_channels_condition", S_CHANNELS)
@pytest.mark.parametrize("dropout_prob", [None])
@pytest.mark.parametrize("multi_query_attention", [False, True])
def test_conditional_gatr_block_equivariance(
    batch_dims,
    num_items,
    num_items_condition,
    mv_channels,
    mv_channels_condition,
    num_heads,
    s_channels,
    s_channels_condition,
    multi_query_attention,
    dropout_prob,
):
    """Tests ConditionalLGATrBlock for equivariance."""
    try:
        net = ConditionalLGATrBlock(
            mv_channels,
            s_channels=s_channels,
            condition_mv_channels=mv_channels_condition,
            condition_s_channels=s_channels_condition,
            attention=SelfAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            crossattention=CrossAttentionConfig(
                num_heads=num_heads,
                multi_query=multi_query_attention,
            ),
            mlp=MLPConfig(),
            dropout_prob=dropout_prob,
        )
    except NotImplementedError:
        # Some features require scalar inputs, and failing without them is fine
        return

    scalars = torch.randn(*batch_dims, num_items, s_channels)
    scalars_condition = torch.randn(*batch_dims, num_items_condition, s_channels_condition)

    data_dims = [
        tuple(list(batch_dims) + [num_items, mv_channels]),
        tuple(list(batch_dims) + [num_items_condition, mv_channels_condition]),
    ]
    check_pin_equivariance(
        net,
        2,
        batch_dims=data_dims,
        fn_kwargs=dict(scalars=scalars, scalars_condition=scalars_condition),
        **MILD_TOLERANCES,
    )



"""Grade dropout."""

import torch

from .linear import grade_project


def grade_dropout(x: torch.Tensor, p: float, training: bool = True) -> torch.Tensor:
    """Multivector dropout, dropping out grades independently.

    Parameters
    ----------
    x : torch.Tensor
        Input data with shape (..., 16).
    p : float
        Dropout probability (assumed the same for each grade).
    training : bool
        Switches between train-time and test-time behaviour.

    Returns
    -------
    outputs : torch.Tensor
        Inputs with dropout applied, shape (..., 16).
    """

    # Project to grades
    x = grade_project(x)

    # Apply standard 1D dropout
    # For whatever reason, that only works with a single batch dimension, so let's reshape a bit
    h = x.view(-1, 5, 16)
    h = torch.nn.functional.dropout1d(h, p=p, training=training, inplace=False)
    h = h.view(x.shape)

    # Combine grades again
    h = torch.sum(h, dim=-2)

    return h



"""Pin-equivariant linear layers between multivector tensors (torch.nn.Modules)."""

import math

import torch
from torch import nn

from ..interface import embed_scalar
from ..primitives.config import gatr_config
from ..primitives.linear import equi_linear


class EquiLinear(nn.Module):
    """Linear layer.

    The forward pass maps multivector inputs with shape (..., in_channels, 16) to multivector
    outputs with shape (..., out_channels, 16) as


        .. code-block::

            outputs[..., j, y] = sum_{i, b, x} weights[j, i, b] basis_map[b, x, y] inputs[..., i, x]
            = sum_i linear(inputs[..., i, :], weights[j, i, :])

    plus an optional bias term for ``outputs[..., :, 0]`` (biases in other multivector components would
    break equivariance). Here ``basis_map`` are precomputed (see ``gatr.primitives.linear``) and weights are the
    learnable weights of this layer. The ``basis_map`` includes 5 elements if the full Lorentz group is
    considered, and 10 elements if only the fully-connected subgroup is considered.
    See the ``use_fully_connected_subgroup`` parameter in ``lgatr.primitives.config.LGATrConfig`` for details.

    If there are auxiliary input scalars, they transform under a linear layer, and mix with the
    scalar components the multivector data. Note that in this layer (and only here) the auxiliary
    scalars are optional.

    This layer supports four initialization schemes:

    - "default": preserves (or actually slightly reducing) the variance of the data in the forward pass
    - "small": variance of outputs is approximately one order of magnitude smaller than for "default"
    - "unit_scalar": outputs will be close to (1, 0, 0, ..., 0)
    - "almost_unit_scalar": similar to "unit_scalar", but with more stochasticity

    We use the "almost_unit_scalar" initialization to preprocess the second
    argument in the ``GeometricBilinears`` layer,
    and the "small" initialization to combine the different attention heads.
    All other linear layers in L-GATr use the "default" initialization.

    Parameters
    ----------
    in_mv_channels : int
        Input multivector channels
    out_mv_channels : int
        Output multivector channels
    bias : bool
        Whether a bias term is added to the scalar component of the multivector outputs
    in_s_channels : int or None
        Input scalar channels. If None, no scalars are expected nor returned.
    out_s_channels : int or None
        Output scalar channels. If None, no scalars are expected nor returned.
    initialization : {"default", "small", "unit_scalar", "almost_unit_scalar"}
            Initialization scheme, see ``EquiLinear`` description for more information.
    """

    def __init__(
        self,
        in_mv_channels: int,
        out_mv_channels: int,
        in_s_channels: int | None = None,
        out_s_channels: int | None = None,
        bias: bool = True,
        initialization: str = "default",
    ) -> None:
        super().__init__()

        # Check inputs
        if initialization in ["unit_scalar", "almost_unit_scalar"]:
            assert bias, "unit_scalar and almost_unit_scalar initialization requires bias"
            if in_s_channels is None:
                raise NotImplementedError(
                    "unit_scalar and almost_unit_scalar initialization is currently only implemented for scalar inputs"
                )

        self._in_mv_channels = in_mv_channels
        self._out_mv_channels = out_mv_channels
        self._in_s_channels = in_s_channels
        self._out_s_channels = out_s_channels
        self._bias = bias

        # MV -> MV
        self.weight = nn.Parameter(
            torch.empty(
                (
                    out_mv_channels,
                    in_mv_channels,
                    gatr_config.num_pin_linear_basis_elements,
                )
            )
        )

        # We only need a separate bias here if that isn't already covered by the linear map from
        # scalar inputs
        self.bias = (
            nn.Parameter(torch.zeros((out_mv_channels, 1)))
            if bias and in_s_channels is None
            else None
        )

        # Scalars -> MV scalars
        self.s2mvs: nn.Linear | None
        mix_factor = 2 if gatr_config.use_fully_connected_subgroup else 1
        if in_s_channels:
            self.s2mvs = nn.Linear(in_s_channels, mix_factor * out_mv_channels, bias=bias)
        else:
            self.s2mvs = None

        # MV scalars -> scalars
        if out_s_channels:
            self.mvs2s = nn.Linear(mix_factor * in_mv_channels, out_s_channels, bias=bias)
        else:
            self.mvs2s = None

        # Scalars -> scalars
        if in_s_channels is not None and out_s_channels is not None:
            self.s2s = nn.Linear(
                in_s_channels, out_s_channels, bias=False
            )  # Bias would be duplicate
        else:
            self.s2s = None

        # Initialization
        self.reset_parameters(initialization)

    def forward(
        self, multivectors: torch.Tensor, scalars: torch.Tensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor | None]:
        """Maps input multivectors and scalars using the most general equivariant linear map.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., in_mv_channels, 16)
        scalars : None or torch.Tensor
            Optional input scalars with shape (..., in_s_channels)

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., out_mv_channels, 16)
        outputs_s : None or torch.Tensor
            Output scalars with shape (..., out_s_channels)
        """

        outputs_mv = equi_linear(multivectors, self.weight)  # (..., out_channels, 16)

        if self.bias is not None:
            bias = embed_scalar(self.bias)
            outputs_mv = outputs_mv + bias

        if self.s2mvs is not None and scalars is not None:
            if gatr_config.use_fully_connected_subgroup:
                outputs_mv[..., [0, -1]] += self.s2mvs(scalars).view(
                    *outputs_mv.shape[:-2], outputs_mv.shape[-2], 2
                )
            else:
                outputs_mv[..., 0] += self.s2mvs(scalars)

        if self.mvs2s is not None:
            if gatr_config.use_fully_connected_subgroup:
                outputs_s = self.mvs2s(multivectors[..., [0, -1]].flatten(start_dim=-2))
            else:
                outputs_s = self.mvs2s(multivectors[..., 0])
            if self.s2s is not None and scalars is not None:
                outputs_s = outputs_s + self.s2s(scalars)
        else:
            outputs_s = None

        return outputs_mv, outputs_s

    def reset_parameters(
        self,
        initialization: str,
        gain: float = 1.0,
        additional_factor=None,
    ) -> None:
        """Initializes the weights of the linear layer.

        We following the initialization philosophy of `Kaiming <https://arxiv.org/pdf/1502.01852>`_,
        ensuring that the variance of the activations during the forward pass is preserved.
        Our implementation deviates from the `torch.nn.Linear default initialization <https://github.com/pytorch/pytorch/blob/v2.7.0/torch/nn/modules/linear.py#L114>`_
        to take the communication between scalar and multivector channels in our linear layer into account.
        For more information, see inline comments in the code.

        Parameters
        ----------
        initialization : {"default", "small", "unit_scalar", "almost_unit_scalar"}
            Initialization scheme, see ``EquiLinear`` description for more information.
        gain : float
            Gain factor for the activations. Should be 1.0 if previous layer has no activation,
            sqrt(2) if it has a ReLU activation, and so on. Can be computed with
            `torch.nn.init.calculate_gain()`.
        additional_factor : float
            Empirically, it has been found that slightly *decreasing* the data variance at each
            layer gives a better performance. In particular, the PyTorch default initialization uses
            an additional factor of 1/sqrt(3) (cancelling the factor of sqrt(3) that naturally
            arises when computing the bounds of a uniform initialization). A discussion of this was
            (to the best of our knowledge) never published, but see
            https://github.com/pytorch/pytorch/issues/57109 and
            https://soumith.ch/files/20141213_gplus_nninit_discussion.htm.
        """
        additional_factor = 1 / math.sqrt(3.0) if additional_factor is None else additional_factor

        # Prefactors depending on initialization scheme
        (
            mv_component_factors,
            mv_factor,
            mvs_bias_shift,
            s_factor,
        ) = self._compute_init_factors(
            initialization,
            gain,
            additional_factor,
        )

        # Following He et al, 1502.01852, we aim to preserve the variance in the forward pass.
        # A sufficient criterion for this is that the variance of the weights is given by
        # `Var[w] = gain^2 / fan`.
        # Here `gain^2` is 2 if the previous layer has a ReLU nonlinearity, 1 for the initial layer,
        # and some other value in other situations (we may not care about this too much).
        # More importantly, `fan` is the number of connections: the number of input elements that
        # get summed over to compute each output element.

        # Let us fist consider the multivector outputs.
        self._init_multivectors(mv_component_factors, mv_factor, mvs_bias_shift)

        # Then let's consider the maps to scalars.
        self._init_scalars(s_factor)

    @staticmethod
    def _compute_init_factors(
        initialization,
        gain,
        additional_factor,
    ):
        """Computes prefactors for the initialization.

        See self.reset_parameters().
        """

        assert initialization in [
            "default",
            "small",
            "unit_scalar",
            "almost_unit_scalar",
        ]

        if initialization == "default":
            mv_factor = gain * additional_factor * math.sqrt(3)
            s_factor = gain * additional_factor * math.sqrt(3)
            mvs_bias_shift = 0.0
        elif initialization == "small":
            # Change scale by a factor of 0.1 in this layer
            mv_factor = 0.1 * gain * additional_factor * math.sqrt(3)
            s_factor = 0.1 * gain * additional_factor * math.sqrt(3)
            mvs_bias_shift = 0.0
        elif initialization == "unit_scalar":
            # Change scale by a factor of 0.1 for MV outputs, and initialize bias around 1
            mv_factor = 0.1 * gain * additional_factor * math.sqrt(3)
            s_factor = gain * additional_factor * math.sqrt(3)
            mvs_bias_shift = 1.0
        elif initialization == "almost_unit_scalar":
            # Change scale by a factor of 0.5 for MV outputs, and initialize bias around 1
            mv_factor = 0.5 * gain * additional_factor * math.sqrt(3)
            s_factor = gain * additional_factor * math.sqrt(3)
            mvs_bias_shift = 1.0

        mv_component_factors = torch.ones(gatr_config.num_pin_linear_basis_elements)
        return mv_component_factors, mv_factor, mvs_bias_shift, s_factor

    def _init_multivectors(self, mv_component_factors, mv_factor, mvs_bias_shift):
        """Weight initialization for maps to multivector outputs."""

        # We have
        # `outputs[..., j, y] = sum_{i, b, x} weights[j, i, b] basis_map[b, x, y] inputs[..., i, x]`
        # The basis maps are more or less grade projections, summing over all basis elements
        # corresponds to (almost) an identity map in the GA space. The sum over `b` and `x` thus
        # does not contribute to `fan` substantially. (We may add a small ad-hoc factor later to
        # make up for this approximation.) However, there is still the sum over incoming channels,
        # and thus `fan ~ mv_in_channels`. Assuming (for now) that the previous layer contained a
        # ReLU activation, we finally have the condition `Var[w] = 2 / mv_in_channels`.
        # Since the variance of a uniform distribution between -a and a is given by
        # `Var[Uniform(-a, a)] = a^2/3`, we should set `a = gain * sqrt(3 / mv_in_channels)`.
        # In theory (see docstring).
        fan_in = self._in_mv_channels
        bound = mv_factor / math.sqrt(fan_in)
        for i, factor in enumerate(mv_component_factors):
            nn.init.uniform_(self.weight[..., i], a=-factor * bound, b=factor * bound)

        # Now let's focus on the scalar components of the multivector outputs.
        # If there are only multivector inputs, all is good. But if scalar inputs contribute them as
        # well, they contribute to the output variance as well.
        # In this case, we initialize such that the multivector inputs and the scalar inputs each
        # contribute half to the output variance.
        # We can achieve this by inspecting the basis maps and seeing that only basis element 0
        # contributes to the scalar output. Thus, we can reduce the variance of the corresponding
        # weights to give a variance of 0.5, not 1.
        if self.s2mvs is not None:
            # contribution from scalar -> mv scalar
            bound = mv_component_factors[0] * mv_factor / math.sqrt(fan_in) / math.sqrt(2)
            nn.init.uniform_(self.weight[..., [0]], a=-bound, b=bound)
            if gatr_config.use_fully_connected_subgroup:
                # contribution from scalar -> mv pseudoscalar
                bound = mv_component_factors[-1] * mv_factor / math.sqrt(fan_in) / math.sqrt(2)
                nn.init.uniform_(self.weight[..., [-1]], a=-bound, b=bound)

        # The same holds for the scalar-to-MV map, where we also just want a variance of 0.5.
        # Note: This is not properly extended to scalar and pseudoscalar outputs yet
        if self.s2mvs is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.s2mvs.weight)
            fan_in = max(fan_in, 1)  # Since in theory we could have 0-channel scalar "data"
            bound = mv_component_factors[0] * mv_factor / math.sqrt(fan_in) / math.sqrt(2)
            nn.init.uniform_(self.s2mvs.weight, a=-bound, b=bound)

            # Bias needs to be adapted, as the overall fan in is different (need to account for MV
            # and s inputs) and we may need to account for the unit_scalar initialization scheme
            if self.s2mvs.bias is not None:
                fan_in = (
                    nn.init._calculate_fan_in_and_fan_out(self.s2mvs.weight)[0]
                    + self._in_mv_channels
                )
                bound = mv_component_factors[0] / math.sqrt(fan_in) if fan_in > 0 else 0
                nn.init.uniform_(self.s2mvs.bias, mvs_bias_shift - bound, mvs_bias_shift + bound)

    def _init_scalars(self, s_factor):
        """Weight initialization for maps to multivector outputs."""

        # If both exist, we need to account for overcounting again, and assign each a target a
        # variance of 0.5.
        # Note: This is not properly extended to scalar and pseudoscalar outputs yet
        models = []
        if self.s2s:
            models.append(self.s2s)
        if self.mvs2s:
            models.append(self.mvs2s)
        for model in models:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(model.weight)
            fan_in = max(fan_in, 1)  # Since in theory we could have 0-channel scalar "data"
            bound = s_factor / math.sqrt(fan_in) / math.sqrt(len(models))
            nn.init.uniform_(model.weight, a=-bound, b=bound)
        # Bias needs to be adapted, as the overall fan in is different (need to account for MV and
        # s inputs)
        if self.mvs2s and self.mvs2s.bias is not None:
            fan_in = nn.init._calculate_fan_in_and_fan_out(self.mvs2s.weight)[0]
            if self.s2s:
                fan_in += nn.init._calculate_fan_in_and_fan_out(self.s2s.weight)[0]
            bound = s_factor / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.mvs2s.bias, -bound, bound)



import pytest
import torch

from lgatr.layers.dropout import GradeDropout
from tests.helpers import MILD_TOLERANCES, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", [(10,)])
@pytest.mark.parametrize("p", [0.0, 0.2])
@pytest.mark.parametrize("training", [True, False])
def test_dropout_shape(training, p, batch_dims):
    """Tests GradeDropout for shape correctness."""

    layer = GradeDropout(p=p)
    if training:
        layer.train()
    else:
        layer.eval()

    mv = torch.randn(*batch_dims, 16)
    s = torch.randn(*batch_dims)
    out_mv, out_s = layer(mv, s)

    assert out_mv.shape == mv.shape
    assert out_s.shape == s.shape


@pytest.mark.parametrize("batch_dims", [(10,)])
@pytest.mark.parametrize("training", [True, False])
def test_dropout_trivial_limit(training, batch_dims):
    """Tests that GradeDropout does nothing when p = 0."""

    layer = GradeDropout(p=0.0)
    if training:
        layer.train()
    else:
        layer.eval()

    mv = torch.randn(*batch_dims, 16)
    s = torch.randn(*batch_dims)
    out_mv, out_s = layer(mv, s)

    torch.testing.assert_close(out_mv, mv, **TOLERANCES)
    torch.testing.assert_close(out_s, s, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", [(10,)])
@pytest.mark.parametrize("p", [0.0, 0.2])
@pytest.mark.parametrize("training", [True, False])
def test_dropout_expectation(training, p, batch_dims, num_trials=10000):
    """Tests GradeDropout for correct expectation."""

    layer = GradeDropout(p=p)
    if training:
        layer.train()
    else:
        layer.eval()

    mv = torch.randn(*batch_dims, 16)
    s = torch.randn(*batch_dims)
    out_mv, out_s = layer(
        mv.unsqueeze(0).expand(num_trials, *mv.shape),
        s.unsqueeze(0).expand(num_trials, *s.shape),
    )
    out_mv = out_mv.mean(dim=0)
    out_s = out_s.mean(dim=0)

    torch.testing.assert_close(
        out_mv, mv, **MILD_TOLERANCES
    )  # Over 10k trials we won't get perfect agreement
    torch.testing.assert_close(out_s, s, **MILD_TOLERANCES)


@pytest.mark.parametrize("batch_dims", [(10,)])
@pytest.mark.parametrize("p", [0.0, 0.2])
def test_dropout_equivariance(p, batch_dims):
    """Tests GradeDropout for equivariance."""

    layer = GradeDropout(p=p)
    layer.eval()
    s = torch.randn(*batch_dims)
    check_pin_equivariance(layer, 1, batch_dims=batch_dims, fn_kwargs=dict(scalars=s), **TOLERANCES)



from .constants import BATCH_DIMS, MILD_TOLERANCES, STRICT_TOLERANCES, TOLERANCES
from .equivariance import check_pin_equivariance, check_pin_invariance
from .geometric_algebra import (
    check_consistence_with_geometric_product,
    check_consistence_with_grade_involution,
    check_consistence_with_reversal,
)



"""Embedding and extracting scalars into multivectors."""

import torch


def embed_scalar(scalars: torch.Tensor) -> torch.Tensor:
    """Embeds a scalar tensor into multivectors.

    Parameters
    ----------
    scalars: torch.Tensor
        Scalar inputs with shape (..., 1).

    Returns
    -------
    multivectors: torch.Tensor
        Multivector outputs with shape (..., 16).
        ``multivectors[..., [0]]`` is the same as ``scalars``. The other components are zero.
    """
    assert scalars.shape[-1] == 1
    non_scalar_shape = list(scalars.shape[:-1]) + [15]
    non_scalar_components = torch.zeros(
        non_scalar_shape, device=scalars.device, dtype=scalars.dtype
    )
    embedding = torch.cat((scalars, non_scalar_components), dim=-1)

    return embedding


def extract_scalar(multivectors: torch.Tensor) -> torch.Tensor:
    """Extracts scalar components from multivectors.

    Parameters
    ----------
    multivectors: torch.Tensor
        Multivector inputs with shape (..., 16).

    Returns
    -------
    scalars: torch.Tensor
        Scalar component of multivectors with shape (..., 1).
    """

    return multivectors[..., [0]]



import pytest
import torch

from lgatr.layers.mlp.geometric_bilinears import GeometricBilinear
from lgatr.primitives.config import gatr_config
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("in_mv_channels", [8])
@pytest.mark.parametrize("out_mv_channels", [10])
@pytest.mark.parametrize("in_s_channels", [3])
@pytest.mark.parametrize("out_s_channels", [5])
@pytest.mark.parametrize("use_bivector", [True, False])
def test_geometric_bilinears_equivariance(
    batch_dims,
    in_mv_channels,
    out_mv_channels,
    in_s_channels,
    out_s_channels,
    use_bivector,
):
    """Tests GeometricBilinear() for equivariance."""
    gatr_config.use_bivector = use_bivector

    layer = GeometricBilinear(
        in_mv_channels,
        out_mv_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
    )
    data_dims = tuple(list(batch_dims) + [in_mv_channels])
    scalars = torch.randn(*batch_dims, in_s_channels)

    check_pin_equivariance(
        layer,
        1,
        fn_kwargs=dict(scalars=scalars),
        batch_dims=data_dims,
        **TOLERANCES,
    )



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any


@dataclass
class MLPConfig:
    """Geometric MLP configuration.

    Parameters
    ----------
    activation : {"relu", "sigmoid", "gelu", "silu}
        Which (gated) activation function to use.
    increase_hidden_channels : int
        Factor by which to increase the number of hidden channels (both multivectors and scalars).
        Vanilla transformers use 4, we use 2 for backward compatibility.
    num_hidden_layers : int
        Number of hidden layers to create.

    Parameters auto-set by LGATr
    ----------------------------
    mv_channels : int
        Number of input multivector channels.
    s_channels : int
        Number of input scalar channels.
    dropout_prob : float or None
        Dropout probability
    """

    mv_channels: int | None = None
    s_channels: int | None = None
    dropout_prob: float | None = None
    activation: str = "gelu"
    increase_hidden_channels: int = 4
    num_hidden_layers: int = 1

    @classmethod
    def cast(cls, config: Any) -> MLPConfig:
        """Casts an object as MLPConfig."""
        if isinstance(config, MLPConfig):
            return config
        if isinstance(config, Mapping):
            return cls(**config)
        raise ValueError(f"Can not cast {config} to {cls}")



import pytest
import torch
from torch import Tensor

from lgatr.utils.misc import minimum_autocast_precision


# Choose dtypes to work on most devices -- torch.bfloat16 is not available on some GPUs
@pytest.mark.parametrize("device,amp_dtype", [("cpu", torch.bfloat16)])
def test_minimum_autocast_precision_inputs(device, amp_dtype):
    """Tests that minimum_autocast_precision() casts inputs correctly"""

    # We wrap a function that just returns the dtypes of the inputs
    @minimum_autocast_precision(torch.float32)
    def return_input_dtypes(*args, **kwargs):
        dtypes = [arg.dtype if isinstance(arg, Tensor) else None for arg in args]
        dtypes += [arg.dtype if isinstance(arg, Tensor) else None for arg in kwargs.values()]
        return dtypes

    # Inputs
    input_dtypes = [
        amp_dtype,
        torch.float32,
        torch.float64,
        torch.int8,
        torch.int32,
        torch.bool,
        None,
    ]
    inputs = [
        "banana" if dtype is None else torch.empty(3, 5, device=device, dtype=dtype)
        for dtype in input_dtypes
    ]
    expected_dtypes = [
        torch.float32,
        torch.float32,
        torch.float64,
        torch.int8,
        torch.int32,
        torch.bool,
        None,
    ]

    # Test that without autocast, nothing happens
    dtypes0 = return_input_dtypes(*inputs)
    for got, expected in zip(dtypes0, input_dtypes, strict=False):
        assert got == expected

    # Test that when autocasting, inputs are correctly casted
    with torch.autocast(device, amp_dtype, enabled=True):
        dtypes1 = return_input_dtypes(*inputs)

    for got, expected in zip(dtypes1, expected_dtypes, strict=False):
        assert got == expected


# Choose dtypes to work on most devices -- torch.bfloat16 is not available on some GPUs
@pytest.mark.parametrize(
    "output_mode,expected_dtype",
    [
        (None, torch.float64),
        (torch.float64, torch.float64),
        ("low", torch.bfloat16),
        ("high", torch.float64),
    ],
)
def test_minimum_autocast_precision_outputs(
    output_mode, expected_dtype, device="cpu", amp_dtype=torch.bfloat16
):
    """Tests that minimum_autocast_precision() casts outputs correctly"""

    # We wrap a function that just returns the dtypes of the inputs
    @minimum_autocast_precision(torch.float32, output=output_mode)
    def sum_(*args):
        outputs = 0.0
        for arg in args:
            outputs = outputs + arg
        return outputs

    # Inputs
    input_dtypes = [torch.bfloat16, torch.float32, torch.float64]
    inputs = [torch.randn((3, 5), device=device, dtype=dtype) for dtype in input_dtypes]

    # Check output dtype
    with torch.autocast(device, amp_dtype, enabled=True):
        outputs = sum_(*inputs)
    assert outputs.dtype == expected_dtype



"""This module provides efficiency improvements over torch's einsum through caching."""

import functools
from collections.abc import Sequence

import opt_einsum
import torch


def custom_einsum(equation: str, *operands: torch.Tensor, path: list[int]) -> torch.Tensor:
    """Computes einsum with a custom contraction order."""

    # Justification: For the sake of performance, we need direct access to torch's private methods.

    return torch._VF.einsum(equation, operands, path=path)


def cached_einsum(equation: str, *operands: torch.Tensor) -> torch.Tensor:
    """Computes einsum with a cached optimal contraction.

    Inspired by upstream
    https://github.com/pytorch/pytorch/blob/v1.13.0/torch/functional.py#L381.
    """
    op_shape = tuple(op.shape for op in operands)
    path = _get_cached_path_for_equation_and_shapes(equation=equation, op_shape=op_shape)

    return custom_einsum(equation, *operands, path=path)


@functools.cache
def _get_cached_path_for_equation_and_shapes(
    equation: str, op_shape: Sequence[torch.Tensor]
) -> list[int]:
    """Provides caching of optimal path."""
    tupled_path = opt_einsum.contract_path(equation, *op_shape, optimize="optimal", shapes=True)[0]

    return [item for pair in tupled_path for item in pair]



"""L-GATr encoder block."""

from dataclasses import replace

import torch
from torch import nn

from .attention import SelfAttention, SelfAttentionConfig
from .layer_norm import EquiLayerNorm
from .mlp.config import MLPConfig
from .mlp.mlp import GeoMLP


class LGATrBlock(nn.Module):
    """L-GATr encoder block.

    Inputs are first processed by a block consisting of LayerNorm, multi-head geometric
    self-attention, and a residual connection. Then the data is processed by a block consisting of
    another LayerNorm, an item-wise two-layer geometric MLP with GeLU activations, and another
    residual connection.

    Parameters
    ----------
    mv_channels : int
        Number of input and output multivector channels
    s_channels: int
        Number of input and output scalar channels
    attention: SelfAttentionConfig
        Attention configuration
    mlp: MLPConfig
        MLP configuration
    dropout_prob : float or None
        Dropout probability
    """

    def __init__(
        self,
        mv_channels: int,
        s_channels: int,
        attention: SelfAttentionConfig,
        mlp: MLPConfig,
        dropout_prob: float | None = None,
    ) -> None:
        super().__init__()

        # Normalization layer (stateless, so we can use the same layer for both normalization instances)
        self.norm = EquiLayerNorm()

        # Self-attention layer
        attention = replace(
            attention,
            in_mv_channels=mv_channels,
            out_mv_channels=mv_channels,
            in_s_channels=s_channels,
            out_s_channels=s_channels,
            output_init="small",
            dropout_prob=dropout_prob,
        )
        self.attention = SelfAttention(attention)

        # MLP block
        mlp = replace(
            mlp,
            mv_channels=mv_channels,
            s_channels=s_channels,
            dropout_prob=dropout_prob,
        )
        self.mlp = GeoMLP(mlp)

    def forward(
        self,
        multivectors: torch.Tensor,
        scalars: torch.Tensor,
        additional_qk_features_mv=None,
        additional_qk_features_s=None,
        **attn_kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass of the transformer encoder block.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., items, mv_channels, 16).
        scalars : torch.Tensor
            Input scalars with shape (..., items, s_channels).
        additional_qk_features_mv : None or torch.Tensor
            Additional multivector Q/K features with shape (..., items, add_qk_mv_channels, 16).
        additional_qk_features_s : None or torch.Tensor
            Additional scalar Q/K features with shape (..., items, add_qk_s_channels, 16).
        **attn_kwargs
            Optional keyword arguments passed to attention.

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., items, mv_channels, 16).
        output_scalars : torch.Tensor
            Output scalars  with shape (..., items, s_channels).
        """

        # Attention block: pre layer norm
        h_mv, h_s = self.norm(multivectors, scalars=scalars)

        # Attention block: self attention
        h_mv, h_s = self.attention(
            h_mv,
            scalars=h_s,
            additional_qk_features_mv=additional_qk_features_mv,
            additional_qk_features_s=additional_qk_features_s,
            **attn_kwargs,
        )

        # Attention block: skip connection
        outputs_mv = multivectors + h_mv
        outputs_s = scalars + h_s

        # MLP block: pre layer norm
        h_mv, h_s = self.norm(outputs_mv, scalars=outputs_s)

        # MLP block: MLP
        h_mv, h_s = self.mlp(h_mv, scalars=h_s)

        # MLP block: skip connection
        outputs_mv = outputs_mv + h_mv
        outputs_s = outputs_s + h_s

        return outputs_mv, outputs_s



from .attention.config import CrossAttentionConfig, SelfAttentionConfig
from .attention.cross_attention import CrossAttention
from .attention.self_attention import SelfAttention
from .conditional_lgatr_block import ConditionalLGATrBlock
from .dropout import GradeDropout
from .layer_norm import EquiLayerNorm
from .lgatr_block import LGATrBlock
from .linear import EquiLinear
from .mlp.config import MLPConfig
from .mlp.geometric_bilinears import GeometricBilinear
from .mlp.mlp import GeoMLP
from .mlp.nonlinearities import ScalarGatedNonlinearity



"""Dynamic attention backend selection."""

from functools import lru_cache
from importlib import metadata

import torch

# common kwargs used in custom attention backends
VARLEN_KWARGS = ["cu_seq_q", "cu_seq_k", "max_q", "max_k"]
XFORMERS_KWARGS = ["attn_bias", "op"]
FLEX_KWARGS = ["score_mod", "block_mask"]
FLASH_KWARGS = ["cu_seqlens_q", "cu_seqlens_k", "max_seqlen_q", "max_seqlen_k"]


@lru_cache
def get_device():
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    return device


_REGISTRY = {}
for ep in metadata.entry_points(group="lgatr.primitives.attention_backends"):
    try:
        # check if entry point code be loaded without ImportError
        module = ep.load()
    except ImportError:
        continue

    if ep.name in ["xformers", "flash"] and get_device() == torch.device("cpu"):
        # xformers and flash-attn are not available on CPU
        continue
    _REGISTRY[ep.name] = module


def get_attention_backend(**kwargs):
    """
    Dynamically determine the attention backend based on the extra keyword arguments.

    Implemented backends:
    - PyTorch's native attention: torch.nn.functional.scaled_dot_product_attention
    - PyTorch's varlen attention: torch.nn.attention.varlen.varlen_attn
    - Xformers attention: xformers.ops.memory_efficient_attention
    - PyTorch's flex_attention: torch.nn.attention.flex_attention.flex_attention
    - Original flash attention (supports variable sequence length): flash_attn.flash_attn_varlen_func
    """
    # check if backend is explicitly specified
    backend = kwargs.get("backend", None)
    if backend in _REGISTRY:
        return _REGISTRY[backend].attention

    # automatic fall-back based on other **kwargs
    if any(kwargs.get(kwarg, None) is not None for kwarg in VARLEN_KWARGS):
        return _REGISTRY["varlen"].attention
    elif any(kwargs.get(kwarg, None) is not None for kwarg in XFORMERS_KWARGS):
        return _REGISTRY["xformers"].attention
    elif any(kwargs.get(kwarg, None) is not None for kwarg in FLEX_KWARGS):
        return _REGISTRY["flex"].attention
    elif any(kwargs.get(kwarg, None) is not None for kwarg in FLASH_KWARGS):
        return _REGISTRY["flash"].attention

    # fall-back to native torch attention
    try:
        return _REGISTRY["native"].attention
    except KeyError as err:
        raise RuntimeError(
            f"No attention backend could be resolved. Available backends: {list(_REGISTRY)}"
        ) from err



import opt_einsum
import pytest
import torch

from lgatr.utils.einsum import cached_einsum

_DIM = 5


@pytest.fixture(name="einsum_eq")
def einsum_eq_fixture() -> str:
    """Provides a non-trivial einsum equation."""
    # Example equation from https://optimized-einsum.readthedocs.io/en/stable/index.html
    return "pi,qj,ijkl,rk,sl->pqrs"


@pytest.fixture(name="example_operands")
def example_operands_fixture() -> tuple[torch.Tensor, ...]:
    """Provides tensors for a non-trivial einsum equation."""
    # Example tensors from https://optimized-einsum.readthedocs.io/en/stable/index.html
    A = torch.rand(_DIM, _DIM, _DIM, _DIM)
    C = torch.rand(_DIM, _DIM)
    return (C, C, A, C, C)


def test_opt_einsum_shape() -> None:
    """Verifies contract_path can be called with shapes.

    Note that we test upstream functionality here, which should not be necessary.
    However, with the latest release (3.3.0), this failed.

    NB: Test does not fail on every example (e.g., does not fail on `einsum_eq_fixture`).
    We use an equation from a failing test (test_pin).
    """
    einsum_eq = "i j k, ... j, ... k -> ... i"
    sizes = (_DIM, _DIM, _DIM)
    operands = tuple(torch.rand(sizes) for _ in sizes)
    op_shape = tuple(x.size() for x in operands)
    assert (
        opt_einsum.contract_path(einsum_eq, *operands, optimize="optimal")[0]
        == opt_einsum.contract_path(einsum_eq, *op_shape, optimize="optimal", shapes=True)[0]
    )


def test_e2e_cached_path(einsum_eq: str, example_operands: tuple[torch.Tensor]) -> None:
    """Checks that torch.einsum and cached_einsum deliver the same values on a non-trivial einsum
    equation."""
    expected_result = torch.einsum(einsum_eq, *example_operands)
    result_with_uncached_path = cached_einsum(einsum_eq, *example_operands)
    result_with_cached_path = cached_einsum(einsum_eq, *example_operands)

    torch.testing.assert_close(expected_result, result_with_uncached_path)
    torch.testing.assert_close(expected_result, result_with_cached_path)



"""Utility functions to test callables for equivariance with respect to SO(1,3)."""

import torch
from lloca.utils.rand_transforms import rand_lorentz


def check_invariance(
    function,
    fn_kwargs=None,
    batch_dims=(1,),
    num_args=1,
    num_checks=3,
    **kwargs,
):
    if fn_kwargs is None:
        fn_kwargs = {}
    if num_args == 1:
        batch_dims = [batch_dims]

    for _ in range(num_checks):
        inputs = [torch.randn(*batch_dims[i], 4) for i in range(num_args)]
        trafo = rand_lorentz((1,) * len(batch_dims))

        outputs = function(*inputs, **fn_kwargs)[0]

        inputs_transformed = [torch.einsum("...ij,...j->...i", trafo, inp) for inp in inputs]
        outputs_of_transformed = function(*inputs_transformed, **fn_kwargs)[0]

        torch.testing.assert_close(outputs, outputs_of_transformed, **kwargs)


def check_equivariance(
    function,
    fn_kwargs=None,
    batch_dims=(1,),
    num_args=1,
    num_checks=3,
    **kwargs,
):
    if fn_kwargs is None:
        fn_kwargs = {}
    if num_args == 1:
        batch_dims = [batch_dims]

    for _ in range(num_checks):
        inputs = [torch.randn(*batch_dims[i], 4) for i in range(num_args)]
        trafo = rand_lorentz((1,) * len(batch_dims))

        outputs = function(*inputs, **fn_kwargs)[0]
        outputs_transformed = torch.einsum("...ij,...j->...i", trafo, outputs)

        inputs_transformed = [torch.einsum("...ij,...j->...i", trafo, inp) for inp in inputs]
        outputs_of_transformed = function(*inputs_transformed, **fn_kwargs)[0]

        torch.testing.assert_close(outputs_transformed, outputs_of_transformed, **kwargs)



"""Linear operations on multivectors, in particular linear basis maps."""

from functools import lru_cache
from pathlib import Path

import torch

from ..utils.einsum import cached_einsum, custom_einsum
from .config import gatr_config

DEFAULT_DEVICE = torch.device("cpu")
DEFAULT_DTYPE = torch.float32


@lru_cache
def _compute_pin_equi_linear_basis(
    use_fully_connected_subgroup: bool = True,
    device=DEFAULT_DEVICE,
    dtype=DEFAULT_DTYPE,
) -> torch.Tensor:
    """Constructs basis elements for Lorentz-equivariant linear maps between multivectors.

    This function is cached.

    Parameters
    ----------
    use_fully_connected_subgroup : bool
        If True, model is only equivariant with respect to
        the fully connected subgroup of the Lorentz group,
        the proper orthochronous Lorentz group :math:`SO^+(1,3)`,
        which does not include parity and time reversal.
        This setting affects how the EquiLinear maps work:
        For :math:`SO^+(1,3)`, they include transitions scalars/pseudoscalars
        vectors/axialvectors and among bivectors, effectively
        treating the pseudoscalar/axialvector representations
        like another scalar/vector.
        Defaults to False, because parity-odd representations
        are usually not important in high-energy physics simulations.
    device : torch.device
        Device
    dtype : torch.dtype
        Dtype

    Returns
    -------
    basis : torch.Tensor
        Basis elements for equivariant linear maps with shape (NUM_PIN_LINEAR_BASIS_ELEMENTS, 16, 16),
        with NUM_PIN_LINEAR_BASIS_ELEMENTS=5 for the full Lorentz group (including the discrete
        operations of parity and time reversal) and 10 for the fully connected subgroup.
    """

    if device not in [DEFAULT_DEVICE, "cpu"] and dtype != DEFAULT_DTYPE:
        basis = _compute_pin_equi_linear_basis(
            use_fully_connected_subgroup=use_fully_connected_subgroup
        )
    else:
        file = (
            "linear_basis_subgroup.pt" if use_fully_connected_subgroup else "linear_basis_full.pt"
        )
        filename = Path(__file__).parent.resolve() / file
        basis = torch.load(filename).to(DEFAULT_DTYPE).to_dense()
    return basis.to(device=device, dtype=dtype)


@lru_cache
def _compute_reversal(device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE) -> torch.Tensor:
    """Constructs a matrix that computes multivector reversal.

    Parameters
    ----------
    device : torch.device
        Device
    dtype : torch.dtype
        Dtype

    Returns
    -------
    reversal_diag : torch.Tensor
        The diagonal of the reversal matrix with shape (16,), consisting of +1 and -1 entries.
    """
    reversal_flat = torch.ones(16, device=device, dtype=dtype)
    reversal_flat[5:15] = -1
    return reversal_flat


@lru_cache
def _compute_grade_involution(device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE) -> torch.Tensor:
    """Constructs a matrix that computes multivector grade involution.

    Parameters
    ----------
    device : torch.device
        Device
    dtype : torch.dtype
        Dtype

    Returns
    -------
    involution_diag : torch.Tensor
        The diagonal of the involution matrix with shape (16,), consisting of +1 and -1 entries.
    """
    involution_flat = torch.ones(16, device=device, dtype=dtype)
    involution_flat[1:5] = -1
    involution_flat[11:15] = -1
    return involution_flat


def equi_linear(x: torch.Tensor, coeffs: torch.Tensor) -> torch.Tensor:
    """Pin-equivariant linear map ``f(x) = sum_{a,j} coeffs_a W^a_ij x_j``.

    The W^a are seven pre-defined basis elements.

    Parameters
    ----------
    x : torch.Tensor
        Input multivector with shape (..., in_channels, 16).
        Batch dimensions must be broadcastable between x and coeffs.
    coeffs : torch.Tensor
        Coefficients for the basis elements with shape (out_channels, in_channels, 10).
        Batch dimensions must be broadcastable between x and coeffs.

    Returns
    -------
    outputs : torch.Tensor
        Result with shape (..., 16).
        Batch dimensions are result of broadcasting between x and coeffs.
    """
    basis = _compute_pin_equi_linear_basis(
        gatr_config.use_fully_connected_subgroup, device=x.device, dtype=x.dtype
    )
    return custom_einsum("y x a, a i j, ... x j -> ... y i", coeffs, basis, x, path=[0, 1, 0, 1])


def grade_project(x: torch.Tensor) -> torch.Tensor:
    """Projects an input tensor to the individual grades.

    The return value is a single tensor with a new grade dimension.

    Parameters
    ----------
    x : torch.Tensor
        Input multivector with shape (..., 16).

    Returns
    -------
    outputs : torch.Tensor
        Output multivector with shape (..., 5, 16).
        The second-to-last dimension indexes the grades.
    """

    # Select kernel on correct device
    basis = _compute_pin_equi_linear_basis(
        gatr_config.use_fully_connected_subgroup,
        device=x.device,
        dtype=x.dtype,
    )

    # First five basis elements are grade projections
    basis = basis[:5]

    # Project to grades
    projections = cached_einsum("g i j, ... j -> ... g i", basis, x)

    return projections


def reverse(x: torch.Tensor) -> torch.Tensor:
    """Computes the reversal of a multivector.

    The reversal has the same scalar, vector, and pseudoscalar components, but flips sign in the
    bivector and trivector components.

    Parameters
    ----------
    x : torch.Tensor
        Input multivector with shape (..., 16).

    Returns
    -------
    outputs : torch.Tensor
        Output multivector with shape (..., 16).
    """
    return _compute_reversal(device=x.device, dtype=x.dtype) * x


def grade_involute(x: torch.Tensor) -> torch.Tensor:
    """Computes the grade involution of a multivector.

    The reversal has the same scalar, bivector, and pseudoscalar components, but flips sign in the
    vector and trivector components.

    Parameters
    ----------
    x : torch.Tensor
        Input multivector with shape (..., 16).

    Returns
    -------
    outputs : torch.Tensor
        Output multivector with shape (..., 16).
    """

    return _compute_grade_involution(device=x.device, dtype=x.dtype) * x



"""Gated nonlinearity on multivectors."""

import torch
from torch import nn

from ...primitives.nonlinearities import (
    gated_gelu,
    gated_relu,
    gated_sigmoid,
    gated_silu,
)


class ScalarGatedNonlinearity(nn.Module):
    """Gated nonlinearity on multivectors.

    Given multivector input x, computes ``f(x_0) * x``, where f can either be ReLU, sigmoid, or GeLU.

    Auxiliary scalar inputs are simply processed with ReLU, sigmoid, GeLU, or SiLU, without gating.

    Parameters
    ----------
    nonlinearity : {"relu", "sigmoid", "gelu", "silu"}
        Non-linearity type
    """

    def __init__(self, nonlinearity: str = "relu") -> None:
        super().__init__()

        gated_fn_dict = dict(
            relu=gated_relu, gelu=gated_gelu, sigmoid=gated_sigmoid, silu=gated_silu
        )
        scalar_fn_dict = dict(
            relu=nn.functional.relu,
            gelu=nn.functional.gelu,
            sigmoid=nn.functional.sigmoid,
            silu=nn.functional.silu,
        )
        try:
            self.gated_nonlinearity = gated_fn_dict[nonlinearity]
            self.scalar_nonlinearity = scalar_fn_dict[nonlinearity]
        except KeyError as exc:
            raise ValueError(
                f"Unknown nonlinearity {nonlinearity} for options {list(gated_fn_dict.keys())}"
            ) from exc

    def forward(
        self, multivectors: torch.Tensor, scalars: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Computes ``f(x_0) * x`` for multivector x, where f is GELU, ReLU, sigmoid, or SiLU.

        f is chosen depending on self.gated_nonlinearity and self.scalar_nonlinearity.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., 16)
        scalars : None or torch.Tensor
            Input scalars with shape (...)

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors with shape (..., 16)
        output_scalars : torch.Tensor
            Output scalars with shape (...)
        """

        gates = multivectors[..., [0]]
        outputs_mv = self.gated_nonlinearity(multivectors, gates=gates)
        outputs_s = self.scalar_nonlinearity(scalars) if scalars is not None else None

        return outputs_mv, outputs_s



"""Equivariant conditional transformer for multivector data."""

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from ..layers import (
    ConditionalLGATrBlock,
    CrossAttentionConfig,
    EquiLinear,
    SelfAttentionConfig,
)
from ..layers.mlp.config import MLPConfig


class ConditionalLGATr(nn.Module):
    """Conditional L-GATr network.
    Assumes that the condition is already preprocessed, e.g. with a non-conditional `LGATr` network.

    It combines num_blocks conditional L-GATr transformer blocks, each consisting of geometric self-attention
    layers, geometric cross-attention layers, a geometric MLP, residual connections, and normalization layers.
    In addition, there are initial and final equivariant linear layers.

    Parameters
    ----------
    num_blocks : int
        Number of transformer blocks.
    in_mv_channels : int
        Number of input multivector channels.
    condition_mv_channels : int
        Number of condition multivector channels.
    out_mv_channels : int
        Number of output multivector channels.
    hidden_mv_channels : int
        Number of hidden multivector channels.
    in_s_channels : None or int
        If not None, sets the number of scalar input channels.
    condition_s_channels : None or int
        If not None, sets the number of scalar condition channels.
    out_s_channels : None or int
        If not None, sets the number of scalar output channels.
    hidden_s_channels : None or int
        If not None, sets the number of scalar hidden channels.
    attention: Dict
        Data for SelfAttentionConfig.
    crossattention: Dict
        Data for CrossAttentionConfig.
    mlp: Dict
        Data for MLPConfig.
    dropout_prob : float or None
        Dropout probability.
    checkpoint_blocks : bool
        Whether to use checkpointing for the transformer blocks to save memory.
    """

    def __init__(
        self,
        num_blocks: int,
        in_mv_channels: int,
        condition_mv_channels: int,
        out_mv_channels: int,
        hidden_mv_channels: int,
        in_s_channels: int | None,
        condition_s_channels: int | None,
        out_s_channels: int | None,
        hidden_s_channels: int | None,
        attention: SelfAttentionConfig,
        crossattention: CrossAttentionConfig,
        mlp: MLPConfig,
        dropout_prob: float | None = None,
        checkpoint_blocks: bool = False,
    ) -> None:
        super().__init__()

        self.linear_in = EquiLinear(
            in_mv_channels,
            hidden_mv_channels,
            in_s_channels=in_s_channels,
            out_s_channels=hidden_s_channels,
        )

        attention = SelfAttentionConfig.cast(attention)
        crossattention = CrossAttentionConfig.cast(crossattention)
        mlp = MLPConfig.cast(mlp)

        self.blocks = nn.ModuleList(
            [
                ConditionalLGATrBlock(
                    mv_channels=hidden_mv_channels,
                    s_channels=hidden_s_channels,
                    condition_mv_channels=condition_mv_channels,
                    condition_s_channels=condition_s_channels,
                    attention=attention,
                    crossattention=crossattention,
                    mlp=mlp,
                    dropout_prob=dropout_prob,
                )
                for _ in range(num_blocks)
            ]
        )
        self.linear_out = EquiLinear(
            hidden_mv_channels,
            out_mv_channels,
            in_s_channels=hidden_s_channels,
            out_s_channels=out_s_channels,
        )
        self._checkpoint_blocks = checkpoint_blocks

    def forward(
        self,
        multivectors: torch.Tensor,
        multivectors_condition: torch.Tensor,
        scalars: torch.Tensor | None = None,
        scalars_condition: torch.Tensor | None = None,
        attn_kwargs=None,
        crossattn_kwargs=None,
    ) -> tuple[torch.Tensor, torch.Tensor | None]:
        """Forward pass of the network.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., items, in_mv_channels, 16).
        multivectors_condition : torch.Tensor
            Input condition multivectors with shape (..., items, in_mv_channels, 16).
        scalars : None or torch.Tensor
            Optional input scalars with shape (..., items, in_s_channels).
        scalars_condition : None or torch.Tensor
            Optional input scalars  with shape (..., items, in_s_channels).
        attn_kwargs: None or torch.Tensor or AttentionBias
            Optional attention arguments.
        crossattn_kwargs: None or torch.Tensor or AttentionBias
            Optional attention arguments for the condition.

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors  with shape (..., items, out_mv_channels, 16).
        outputs_s : None or torch.Tensor
            Output scalars with shape (..., items, out_s_channels). None if out_s_channels=None.
        """
        attn_kwargs = attn_kwargs if attn_kwargs is not None else {}
        crossattn_kwargs = crossattn_kwargs if crossattn_kwargs is not None else {}

        # Decode condition into main track with
        h_mv, h_s = self.linear_in(multivectors, scalars=scalars)
        for block in self.blocks:
            if self._checkpoint_blocks:
                h_mv, h_s = checkpoint(
                    block,
                    h_mv,
                    use_reentrant=False,
                    scalars=h_s,
                    multivectors_condition=multivectors_condition,
                    scalars_condition=scalars_condition,
                    attn_kwargs=attn_kwargs,
                    crossattn_kwargs=crossattn_kwargs,
                )
            else:
                h_mv, h_s = block(
                    h_mv,
                    scalars=h_s,
                    multivectors_condition=multivectors_condition,
                    scalars_condition=scalars_condition,
                    attn_kwargs=attn_kwargs,
                    crossattn_kwargs=crossattn_kwargs,
                )

        outputs_mv, outputs_s = self.linear_out(h_mv, scalars=h_s)

        return outputs_mv, outputs_s



from importlib.metadata import version as _pkg_version

from .interface.axialvector import embed_axialvector, extract_axialvector
from .interface.pseudoscalar import embed_pseudoscalar, extract_pseudoscalar
from .interface.scalar import embed_scalar, extract_scalar
from .interface.spurions import get_num_spurions, get_spurions
from .interface.vector import embed_vector, extract_vector
from .layers.attention.config import CrossAttentionConfig, SelfAttentionConfig
from .layers.mlp.config import MLPConfig
from .nets.conditional_lgatr import ConditionalLGATr
from .nets.lgatr import LGATr
from .nets.lgatr_slim import LGATrSlim
from .primitives.config import gatr_config

__version__ = _pkg_version("lgatr")



"""PyTorch's modern and flexible flex_attention backend."""

try:
    from torch.nn.attention.flex_attention import flex_attention
except ModuleNotFoundError as err:
    raise ImportError(
        "torch>=2.5 is not installed. Run 'pip install lgatr[flex-attention]'."
    ) from err

attention = flex_attention



"""MLP with geometric product."""

import torch
from torch import nn

from ...primitives.config import gatr_config
from ..dropout import GradeDropout
from ..linear import EquiLinear
from .config import MLPConfig
from .geometric_bilinears import GeometricBilinear
from .nonlinearities import ScalarGatedNonlinearity


class GeoMLP(nn.Module):
    """MLP with geometric product.

    This is a core component of L-GATr's transformer blocks. It is similar to a regular MLP, except
    that it uses geometric bilinears (the geometric product) in place of the first linear layer.

    Assumes input has shape (..., channels, 16), output has shape (..., channels, 16),
    will create hidden layers with shape (..., increase_hidden_channels*channels, 16).

    Parameters
    ----------
    config: MLPConfig
        Configuration object
    """

    def __init__(
        self,
        config: MLPConfig,
    ) -> None:
        super().__init__()

        # Store settings
        self.config = config

        assert config.mv_channels is not None
        s_channels = None if config.s_channels is None else config.s_channels

        mv_channels_list = [config.mv_channels]
        mv_channels_list.extend(
            [config.increase_hidden_channels * config.mv_channels] * config.num_hidden_layers
        )
        mv_channels_list.append(config.mv_channels)
        if s_channels is not None:
            s_channels_list = [s_channels]
            s_channels_list.extend(
                [config.increase_hidden_channels * s_channels] * config.num_hidden_layers
            )
            s_channels_list.append(s_channels)
        else:
            s_channels_list = [None] * (len(mv_channels_list))

        layers: list[nn.Module] = []

        if config.num_hidden_layers >= 0:
            kwargs = dict(
                in_mv_channels=mv_channels_list[0],
                out_mv_channels=mv_channels_list[1],
                in_s_channels=s_channels_list[0],
                out_s_channels=s_channels_list[1],
            )
            if gatr_config.use_geometric_product:
                layers.append(GeometricBilinear(**kwargs))
            else:
                layers.append(ScalarGatedNonlinearity(config.activation))
                layers.append(EquiLinear(**kwargs))
            if config.dropout_prob is not None:
                layers.append(GradeDropout(config.dropout_prob))

            for in_, out, in_s, out_s in zip(
                mv_channels_list[1:-1],
                mv_channels_list[2:],
                s_channels_list[1:-1],
                s_channels_list[2:],
                strict=False,
            ):
                layers.append(ScalarGatedNonlinearity(config.activation))
                layers.append(EquiLinear(in_, out, in_s_channels=in_s, out_s_channels=out_s))
                if config.dropout_prob is not None:
                    layers.append(GradeDropout(config.dropout_prob))

        self.layers = nn.ModuleList(layers)

    def forward(
        self, multivectors: torch.Tensor, scalars: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor | None]:
        """Forward pass.

        Parameters
        ----------
        multivectors : torch.Tensor
            Input multivectors with shape (..., mv_channels, 16).
        scalars : None or torch.Tensor
            Optional input scalars with shape (..., s_channels).

        Returns
        -------
        outputs_mv : torch.Tensor
            Output multivectors  with shape (..., mv_channels, 16).
        outputs_s : None or torch.Tensor
            Output scalars with shape (..., s_channels).
        """

        mv, s = multivectors, scalars

        for layer in self.layers:
            mv, s = layer(mv, scalars=s)

        return mv, s



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any


@dataclass
class SelfAttentionConfig:
    """Configuration for self-attention.

    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    multi_query: bool
        Whether to do multi-query attention, default is False.
        Multi-query attention decreases memory consumption and parameter count
        by using a single set of keys and values for all heads.
    increase_hidden_channels : int
        Factor by which to increase the number of hidden channels (both multivectors and scalars).
        Vanilla transformers use 1, we use 2 for backward compatibility.
    head_scale: bool
        Whether to use HeadScaleMHA following the NormFormer, see https://arxiv.org/pdf/2110.09456.
        Before combining the heads, each head is scaled by a learnable parameter.


    Parameters auto-set by LGATr
    ----------------------------
    in_mv_channels : int
        Number of input multivector channels.
    out_mv_channels : int
        Number of output multivector channels.
    in_s_channels : int
        Input scalar channels. If None, no scalars are expected nor returned.
    out_s_channels : int
        Output scalar channels. If None, no scalars are expected nor returned.
    additional_qk_mv_channels : int
        Whether additional multivector features for the keys and queries will be provided.
    additional_qk_s_channels : int
        Whether additional scalar features for the keys and queries will be provided.
    output_init : str
        Initialization scheme for final linear layer
    dropout_prob : float or None
        Dropout probability
    """

    in_mv_channels: int | None = None
    out_mv_channels: int | None = None
    in_s_channels: int | None = None
    out_s_channels: int | None = None
    additional_qk_mv_channels: int = 0
    additional_qk_s_channels: int = 0
    output_init: str = "default"
    dropout_prob: float | None = None
    num_heads: int = 8
    multi_query: bool = False
    increase_hidden_channels: int = 1
    head_scale: bool = False

    @property
    def hidden_mv_channels(self) -> int | None:
        """Returns the number of hidden multivector channels."""

        return max(self.increase_hidden_channels * self.in_mv_channels // self.num_heads, 1)

    @property
    def hidden_s_channels(self) -> int | None:
        """Returns the number of hidden scalar channels."""

        if self.in_s_channels is None:
            return None

        hidden_s_channels = max(
            self.increase_hidden_channels * self.in_s_channels // self.num_heads, 4
        )

        return hidden_s_channels

    @classmethod
    def cast(cls, config: Any) -> SelfAttentionConfig:
        """Casts an object as SelfAttentionConfig."""
        if isinstance(config, SelfAttentionConfig):
            return config
        if isinstance(config, Mapping):
            return cls(**config)
        raise ValueError(f"Can not cast {config} to {cls}")


@dataclass
class CrossAttentionConfig:
    """Configuration for cross-attention.

    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    multi_query: bool
        Whether to do multi-query attention, default is False.
        Multi-query attention decreases memory consumption and parameter count
        by using a single set of keys and values for all heads.
    increase_hidden_channels : int
        Factor by which to increase the number of hidden channels (both multivectors and scalars).
        Vanilla transformers use 1, we use 2 for backward compatibility.
    head_scale: bool
        Whether to use HeadScaleMHA following the NormFormer, see https://arxiv.org/pdf/2110.09456.
        Before combining the heads, each head is scaled by a learnable parameter.

    Parameters auto-set by LGATr
    ----------------------------
    in_q_mv_channels : int
        Number of input query multivector channels.
    in_kv_mv_channels : int
        Number of input key/value multivector channels.
    out_mv_channels : int
        Number of output multivector channels.
    in_q_s_channels : int
        Input query scalar channels. If None, no scalars are expected nor returned.
    in_kv_s_channels : int
        Input key/value scalar channels. If None, no scalars are expected nor returned.
    out_s_channels : int
        Output scalar channels. If None, no scalars are expected nor returned.
    additional_q_mv_channels : int
        Whether additional multivector features for the queries will be provided.
    additional_q_s_channels : int
        Whether additional scalar features for the queries will be provided.
    additional_k_mv_channels : int
        Whether additional multivector features for the keys will be provided.
    additional_k_s_channels : int
        Whether additional scalar features for the keys will be provided.
    output_init : str
        Initialization scheme for final linear layer
    dropout_prob : float or None
        Dropout probability
    """

    in_q_mv_channels: int | None = None
    in_kv_mv_channels: int | None = None
    out_mv_channels: int | None = None
    out_s_channels: int | None = None
    in_q_s_channels: int | None = None
    in_kv_s_channels: int | None = None
    additional_q_mv_channels: int = 0
    additional_q_s_channels: int = 0
    additional_k_mv_channels: int = 0
    additional_k_s_channels: int = 0
    output_init: str = "default"
    dropout_prob: float | None = None
    num_heads: int = 8
    multi_query: bool = False
    increase_hidden_channels: int = 1
    head_scale: bool = False

    @property
    def hidden_mv_channels(self) -> int | None:
        """Returns the number of hidden multivector channels."""

        return max(self.increase_hidden_channels * self.in_q_mv_channels // self.num_heads, 1)

    @property
    def hidden_s_channels(self) -> int | None:
        """Returns the number of hidden scalar channels."""

        if self.in_q_s_channels is None:
            assert self.in_kv_s_channels is None
            return None

        return max(self.increase_hidden_channels * self.in_q_s_channels // self.num_heads, 4)

    @classmethod
    def cast(cls, config: Any) -> CrossAttentionConfig:
        """Casts an object as CrossAttentionConfig."""
        if isinstance(config, CrossAttentionConfig):
            return config
        if isinstance(config, Mapping):
            return cls(**config)
        raise ValueError(f"Can not cast {config} to {cls}")



import pytest
import torch

from lgatr.primitives.dropout import grade_dropout
from tests.helpers import (
    BATCH_DIMS,
    MILD_TOLERANCES,
    TOLERANCES,
    check_pin_equivariance,
)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("p", [0.0, 0.2])
@pytest.mark.parametrize("training", [True, False])
def test_dropout_shape(training, p, batch_dims):
    """Tests grade_dropout() map for shape correctness."""
    x = torch.randn(*batch_dims, 16)
    y = grade_dropout(x, p=p, training=training)
    assert y.shape == x.shape


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("training", [False, True])
def test_dropout_trivial_limit(training, batch_dims):
    """Tests that grade_dropout() does nothing when p = 0."""
    x = torch.randn(*batch_dims, 16)
    y = grade_dropout(x, p=0.0, training=training)

    torch.testing.assert_close(x, y, **TOLERANCES)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("p", [0.0, 0.2])
def test_dropout_expectation(p, batch_dims, num_trials=10000):
    """Tests that grade_dropout() has train-time and test-time behaviour that is identical in
    expectation."""
    x = torch.randn(*batch_dims, 16)
    y_train = grade_dropout(x.unsqueeze(0).expand(num_trials, *x.shape), p=p, training=True).mean(
        dim=0
    )
    y_test = grade_dropout(x, p=p, training=False)

    torch.testing.assert_close(y_test, x, **TOLERANCES)
    torch.testing.assert_close(
        y_train, x, **MILD_TOLERANCES
    )  # Over 10k trials we won't get perfect agreement


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("p", [0.0, 0.2])
def test_dropout_equivariance(p, batch_dims):
    """Tests the grade_dropout() primitive for equivariance (at test time)."""
    check_pin_equivariance(
        grade_dropout, 1, batch_dims=batch_dims, fn_kwargs=dict(training=False, p=p), **TOLERANCES
    )



import pytest
import torch

from lgatr.interface import embed_vector, extract_vector
from tests.helpers import BATCH_DIMS, TOLERANCES


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
def test_vector_embedding_consistency(batch_dims):
    """Tests whether Lorentz vector embeddings into multivectors are cycle consistent."""
    vectors = torch.randn(*batch_dims, 4)
    multivectors = embed_vector(vectors)
    vectors_reencoded = extract_vector(multivectors)
    torch.testing.assert_close(vectors, vectors_reencoded, **TOLERANCES)



import pytest
import torch

from lgatr.layers import SelfAttention, SelfAttentionConfig
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize(
    "num_items,in_channels,out_channels,increase_hidden_channels", [(2, 4, 4, 2)]
)
@pytest.mark.parametrize("in_s_channels,out_s_channels", [(17, 13), (11, None)])
@pytest.mark.parametrize("num_heads", [4, 1])
@pytest.mark.parametrize("multi_query,head_scale", [(True, True), (False, False)])
def test_attention_equivariance(
    batch_dims,
    num_items,
    in_channels,
    out_channels,
    num_heads,
    head_scale,
    in_s_channels,
    out_s_channels,
    multi_query,
    increase_hidden_channels,
):
    """Tests the SelfAttention layer for Pin equivariance, with scalar inputs."""

    config = SelfAttentionConfig(
        in_mv_channels=in_channels,
        out_mv_channels=out_channels,
        in_s_channels=in_s_channels,
        out_s_channels=out_s_channels,
        num_heads=num_heads,
        head_scale=head_scale,
        multi_query=multi_query,
        increase_hidden_channels=increase_hidden_channels,
    )
    layer = SelfAttention(config)

    data_dims = tuple(list(batch_dims) + [num_items, in_channels])
    scalars = torch.randn(*batch_dims, num_items, in_s_channels)
    check_pin_equivariance(
        layer,
        1,
        batch_dims=data_dims,
        fn_kwargs=dict(scalars=scalars),
        spin=True,
        **TOLERANCES,
    )



"""Equivariant normalization layers."""

import torch
from torch import nn

from ..primitives import equi_layer_norm


class EquiLayerNorm(nn.Module):
    """Layer normalization.

    Rescales input such that ``mean_channels |inputs|^2 = 1``, where the norm is the GA norm and the
    mean goes over the channel dimensions.

    In addition, the layer performs a regular LayerNorm operation on auxiliary scalar inputs.

    Parameters
    ----------
    mv_channel_dim : int
        Channel dimension index for multivector inputs. Defaults to the second-last entry (last are
        the multivector components).
    epsilon : float
        Small numerical factor to avoid instabilities. We use a reasonably large number to balance
        issues that arise from some multivector components not contributing to the norm.
    """

    def __init__(self, mv_channel_dim=-2, epsilon: float = 0.01):
        super().__init__()
        self.mv_channel_dim = mv_channel_dim
        self.epsilon = epsilon

    def forward(
        self, multivectors: torch.Tensor, scalars: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass. Computes equivariant LayerNorm for multivectors.

        Parameters
        ----------
        multivectors : torch.Tensor
            Multivector inputs with shape (..., 16).
        scalars : torch.Tensor
            Scalar inputs with shape (...).

        Returns
        -------
        outputs_mv : torch.Tensor
            Normalized multivectors with shape (..., 16).
        output_scalars : torch.Tensor
            Normalized scalars with shape (...).
        """

        outputs_mv = equi_layer_norm(
            multivectors, channel_dim=self.mv_channel_dim, epsilon=self.epsilon
        )
        normalized_shape = scalars.shape[-1:]
        outputs_s = torch.nn.functional.layer_norm(scalars, normalized_shape=normalized_shape)

        return outputs_mv, outputs_s



"""xformers memory-efficient attention backend."""

import torch

try:
    from xformers.ops import memory_efficient_attention
except ModuleNotFoundError as err:
    raise ImportError(
        "xformers is not installed. Run 'pip install lgatr[xformers-attention]'."
    ) from err


@torch.compiler.disable()
def attention(query, key, value, dtype=None, **kwargs):
    """Pass to xformers memory-efficient attention.
    Note that this xformers expects the shape (batch, head, items, channel).

    Parameters
    ----------
    query : torch.Tensor
        Queries with shape (batch, head, items_out, channel)
    key : torch.Tensor
        Keys with shape (batch, head, items_in, channel)
    value : torch.Tensor
        Values with shape (batch, head, items_in, channel)
    dtype : torch.dtype, optional
        If specified, cast input tensors to this dtype before passing to attention.
        This can be useful to trigger flash-attention.
    **kwargs
        Additional keyword arguments passed to memory_efficient_attention.

    Returns
    -------
    out : torch.Tensor
        Result with shape (batch, head, items_out, channel)
    """
    assert len(query.shape) == 4, (
        "xformers constrains attention input shape to (batch, head, items, channel)."
    )
    if key.shape[1] != query.shape[1]:
        # manual broadcasting for key and value; required for multi-query attention
        key = key.expand(key.shape[0], query.shape[1], *key.shape[2:])
        value = value.expand(value.shape[0], query.shape[1], *value.shape[2:])

    if dtype is not None:
        in_dtype = query.dtype
        query, key, value = query.to(dtype), key.to(dtype), value.to(dtype)
    else:
        in_dtype = None

    # xformers expects input shape (batch, item, head, channel)
    query = query.transpose(1, 2).contiguous()
    key = key.transpose(1, 2).contiguous()
    value = value.transpose(1, 2).contiguous()

    out = memory_efficient_attention(query, key, value, **kwargs)
    out = out.transpose(1, 2).contiguous()

    if in_dtype is not None:
        out = out.to(in_dtype)
    return out



import pytest
import torch

from lgatr.primitives import sdp_attention
from tests.helpers import BATCH_DIMS, TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("num_s_channels_out", [11, 1])
@pytest.mark.parametrize("num_s_channels_in", [17, 1])
@pytest.mark.parametrize("num_mv_channels_out", [3, 1])
@pytest.mark.parametrize("num_mv_channels_in", [2, 1])
@pytest.mark.parametrize("num_tokens_out", [5, 1])
@pytest.mark.parametrize("num_tokens_in", [7, 1])
def test_scalar_attention_shape(
    batch_dims,
    num_tokens_in,
    num_tokens_out,
    num_mv_channels_in,
    num_mv_channels_out,
    num_s_channels_in,
    num_s_channels_out,
):
    """Tests that outputs of scalar_attention() have correct shape."""
    # Generate inputs
    q_mv = torch.randn(*batch_dims, num_tokens_out, num_mv_channels_in, 16)
    k_mv = torch.randn(*batch_dims, num_tokens_in, num_mv_channels_in, 16)
    v_mv = torch.randn(*batch_dims, num_tokens_in, num_mv_channels_out, 16)
    q_s = torch.randn(*batch_dims, num_tokens_out, num_s_channels_in)
    k_s = torch.randn(*batch_dims, num_tokens_in, num_s_channels_in)
    v_s = torch.randn(*batch_dims, num_tokens_in, num_s_channels_out)

    # Compute attention outputs
    outputs, outputs_scalar = sdp_attention(q_mv, k_mv, v_mv, q_s, k_s, v_s)

    # Check shape of outputs
    assert outputs.shape == (*batch_dims, num_tokens_out, num_mv_channels_out, 16)
    assert outputs_scalar.shape == (*batch_dims, num_tokens_out, num_s_channels_out)


@pytest.mark.parametrize("batch_dims", BATCH_DIMS)
@pytest.mark.parametrize("num_scalars", [5])
@pytest.mark.parametrize("key_dim", [2])
@pytest.mark.parametrize("item_dim", [3])
def test_scalar_attention_equivariance(batch_dims, key_dim, item_dim, num_scalars):
    """Tests scalar_attention() for Pin equivariance."""
    data_dims = tuple(list(batch_dims) + [item_dim, key_dim])
    queries_scalar = torch.randn(*batch_dims, item_dim, num_scalars)
    keys_scalar = torch.randn(*batch_dims, item_dim, num_scalars)
    values_scalar = torch.randn(*batch_dims, item_dim, num_scalars)
    kwargs = dict(q_s=queries_scalar, k_s=keys_scalar, v_s=values_scalar)
    check_pin_equivariance(
        sdp_attention, 3, batch_dims=[data_dims] * 3, fn_kwargs=kwargs, **TOLERANCES
    )



"""Embedding and extracting vectors into multivectors."""

import torch


def embed_vector(vector: torch.Tensor) -> torch.Tensor:
    """Embeds Lorentz vectors in multivectors.

    Parameters
    ----------
    vector : torch.Tensor
        Lorentz vector with shape (..., 4)

    Returns
    -------
    multivector : torch.Tensor
        Embedding into multivector with shape (..., 16).
    """

    # Create multivector tensor with same batch shape, same device, same dtype as input
    batch_shape = vector.shape[:-1]
    multivector = torch.zeros(*batch_shape, 16, dtype=vector.dtype, device=vector.device)

    # Embedding into Lorentz vectors
    multivector[..., 1:5] = vector

    return multivector


def extract_vector(multivector: torch.Tensor) -> torch.Tensor:
    """Given a multivector, extract a Lorentz vector.

    Parameters
    ----------
    multivector : torch.Tensor
        Multivector with shape (..., 16).

    Returns
    -------
    vector : torch.Tensor
        Lorentz vector with shape (..., 4)
    """

    vector = multivector[..., 1:5]

    return vector



"""Invariants, e.g. inner product, absolute squared norm, pin invariants."""

import math
from functools import lru_cache

import torch

from ..utils.einsum import cached_einsum
from ..utils.misc import minimum_autocast_precision
from .linear import DEFAULT_DEVICE, DEFAULT_DTYPE


@lru_cache
def _load_inner_product_factors(device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE) -> torch.Tensor:
    """Constructs an array of 1's and -1's for the metric of the space,
    used to compute the inner product.

    Parameters
    ----------
    device : torch.device
        Device
    dtype : torch.dtype
        Dtype

    Returns
    -------
    ip_factors : torch.Tensor
        Inner product factors with shape (16,)
    """

    _INNER_PRODUCT_FACTORS = [1, 1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, -1, -1]
    factors = torch.tensor(
        _INNER_PRODUCT_FACTORS, dtype=DEFAULT_DTYPE, device=DEFAULT_DEVICE
    ).to_dense()
    return factors.to(device=device, dtype=dtype)


@lru_cache
def _load_metric_grades(device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE) -> torch.Tensor:
    """Generate tensor of the diagonal of the GA metric, combined with a grade projection.

    Parameters
    ----------
    device : torch.device
        Device
    dtype : torch.dtype
        Dtype

    Returns
    -------
    torch.Tensor
        Metric grades with shape (5, 16)
    """
    m = _load_inner_product_factors(device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE)
    m_grades = torch.zeros(5, 16, device=DEFAULT_DEVICE, dtype=DEFAULT_DTYPE)
    offset = 0
    for k in range(4 + 1):
        d = math.comb(4, k)
        m_grades[k, offset : offset + d] = m[offset : offset + d]
        offset += d
    return m_grades.to(device=device, dtype=dtype)


def inner_product(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    """Computes the inner product of multivectors ``f(x,y) = <x, y> = <~x y>_0``.

    Equal to ``geometric_product(reverse(x), y)[..., [0]]`` (but faster).

    Parameters
    ----------
    x : torch.Tensor
        First input multivector with shape (..., 16) or (..., channels, 16).
        Batch dimensions must be broadcastable between x and y.
    y : torch.Tensor
        Second input multivector with shape (..., 16) or (..., channels, 16).
        Batch dimensions must be broadcastable between x and y.

    Returns
    -------
    outputs : torch.Tensor
        Result with shape (..., 1).
        Batch dimensions are result of broadcasting between x and y.
    """

    x = x * _load_inner_product_factors(device=x.device, dtype=x.dtype)

    outputs = cached_einsum("... i, ... i -> ...", x, y)

    # We want the output to have shape (..., 1)
    outputs = outputs.unsqueeze(-1)

    return outputs


@minimum_autocast_precision(torch.float32)
def abs_squared_norm(x: torch.Tensor) -> torch.Tensor:
    """Computes a modified version of the squared norm that is positive semidefinite and can
    therefore be used in layer normalization.

    Parameters
    ----------
    x : torch.Tensor
        Input multivector with shape (..., 16).

    Returns
    -------
    outputs : torch.Tensor
        Geometric algebra norm of x with shape (..., 1).
    """
    m = _load_metric_grades(device=x.device, dtype=x.dtype)
    abs_squared_norms = (
        cached_einsum("... i, ... i, g i -> ... g", x, x, m).abs().sum(-1, keepdim=True)
    )
    return abs_squared_norms



"""Unit tests of nonlinearity primitives."""

import pytest
import torch

from lgatr.primitives import abs_squared_norm, equi_layer_norm
from tests.helpers import TOLERANCES, check_pin_equivariance


@pytest.mark.parametrize("batch_dims", [(7, 9)])
@pytest.mark.parametrize("scale", [0.1, 1.0, 50.0])
def test_equi_layer_norm_correctness(batch_dims, scale):
    """Tests whether the output of equi_layer_norm has the correct variance."""
    inputs = scale * torch.randn(*batch_dims, 16)
    normalized_inputs = equi_layer_norm(inputs, gain=1.0, epsilon=1e-9)
    variance = torch.mean(abs_squared_norm(normalized_inputs))
    torch.testing.assert_close(variance, torch.ones_like(variance), **TOLERANCES)


@pytest.mark.parametrize("batch_dims", [(7, 9)])
def test_equi_layer_norm_equivariance(batch_dims):
    """Tests equi_layer_norm() primitive for equivariance."""
    check_pin_equivariance(equi_layer_norm, 1, batch_dims=batch_dims, **TOLERANCES)



"""Geometric algebra operations based on the clifford library."""

import clifford
import numpy as np
import torch

LAYOUT, BLADES = clifford.Cl(1, 3)


def np_to_mv(array):
    """Shorthand to transform a numpy array to a Pin(1,3) multivector."""
    return clifford.MultiVector(LAYOUT, value=array)


def tensor_to_mv(tensor):
    """Shorthand to transform a numpy array to a Pin(1,3) multivector."""
    return np_to_mv(tensor.detach().cpu().numpy())


def tensor_to_mv_list(tensor):
    """Transforms a torch.Tensor to a list of multivector objects."""

    tensor = tensor.reshape((-1, 16))
    mv_list = [tensor_to_mv(x) for x in tensor]

    return mv_list


def mv_list_to_tensor(multivectors, batch_shape=None):
    """Transforms a list of multivector objects to a torch.Tensor."""

    tensor = torch.from_numpy(np.array([mv.value for mv in multivectors])).to(torch.float32)
    if batch_shape is not None:
        tensor = tensor.reshape(*batch_shape, 16)

    return tensor


def sample_pin_multivector(spin: bool = False, rng: np.random.Generator | None = None):
    """Samples from the Pin(1,3) group as a product of reflections."""

    if rng is None:
        rng = np.random.default_rng()

    # Sample number of reflections we want to multiply
    if spin:
        i = np.random.randint(3) * 2
    else:
        i = np.random.randint(5)

    # If no reflections, just return unit scalar
    if i == 0:
        return BLADES[""]

    multivector = 1.0
    for _ in range(i):
        # Sample reflection vector
        vector = np.zeros(16)
        vector[2:5] = rng.normal(size=3) * 2
        norm = np.linalg.norm(vector[2:5])
        vector[1] = (rng.uniform(size=1)[0] - 0.5) * norm

        vector_mv = np_to_mv(vector)
        vector_mv = vector_mv / abs(vector_mv.mag2()) ** 0.5

        # Multiply together (geometric product)
        multivector = multivector * vector_mv

    return multivector


def get_parity(mv):
    """Gets parity of a clifford multivector.

    Given a clifford multivector, returns True if it is pure-odd-grade, False if it is pure-even
    grade, and raises a RuntimeError if it is mixed.
    """
    if mv == mv.even:
        return False
    if mv == mv.odd:
        return True
    raise RuntimeError(f"Mixed-grade multivector: {mv}")


def sandwich(u, x):
    """Given clifford multivectors, computes their sandwich product.

    Specifically, given a Pin element u and a PGA element x, both given as clifford multivectors,
    computes the sandwich product
    ```
    sandwich(x, u) = (-1)^(grade(u) * grade(x)) u x u^{-1} .
    ```

    If `u` is of odd grades, then this is equal to `u * grade_involute(x) * u^{-1}`.
    If `u` is of even grades, then this is equal to `u * x * u^{-1}`.
    """

    if get_parity(u):
        return u * x.gradeInvol() * u.shirokov_inverse()

    return u * x * u.shirokov_inverse()


class SlowRandomPinTransform:
    """Random Pin transform on a multivector torch.Tensor.

    Slow, only used for testing purposes. Breaks computational graph.
    """

    def __init__(self, spin=False, rng=None):
        super().__init__()
        self._u = sample_pin_multivector(spin, rng)
        self._u_inverse = self._u.shirokov_inverse()

    def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
        """Apply Pin transformation to multivector inputs."""
        # Input shape
        assert inputs.shape[-1] == 16
        batch_dims = inputs.shape[:-1]

        # Convert inputs to list of multivectors
        inputs_mv = tensor_to_mv_list(inputs)

        # Transform
        outputs_mv = [sandwich(self._u, x) for x in inputs_mv]

        # Back to tensor
        outputs = mv_list_to_tensor(outputs_mv, batch_shape=batch_dims)

        return outputs
